/**
  *  \file client/remote.cc
  *  \brief Remote Control Handling
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "client/remote.h"
#include "game/cmds.h"
#include "game/ship.h"
#include "game/team.h"
#include "util/nls.h"
#include "ui/msgbox.h"

/** Remote-control verbs. Indexed by TRemoteControlVerb. */
static const char*const rc_verbs[] = {
    "allow",
    "forbid",
    "drop",
    "control",
};

/** Mapping of TRemoteControlSetting to verb that was used to enter
    this state. In particular, this also means that sending the
    opposite verb leaves the state. */
static const uint8_t rc_verb_index[] = {
    rcv_Forbid,                    // rc_Forbidden
    rcv_Allow,                     // rc_Normal
    rcv_Control,                   // rc_RemoteControlled
    rcv_Control,                   // rc_Applying
    rcv_Drop,                      // rc_Dropping
    rcv_Drop,                      // rc_Other
    rcv_Drop,                      // rc_OtherForbidden
    rcv_Allow,                     // rc_OurRemoteControlled
};

/** Get old remote control state (beginning of turn).
    Returns the logical state of the specified ship. */
TRemoteControlSetting
getOldRemoteControlState(const GShip& sh)
{
    // phost.pas::GetRCFlag
    int rc = sh.getRemoteControlFlag();
    int ow = sh.getOwner();

    /* Special case. Is it used? */
    if (rc == 0)
        rc = ow;

    if (rc < 0) {
        /* Remote control forbidden */
        if (ow == getPlayerId())
            return rc_Forbidden;
        else
            return rc_OtherForbidden;
    } else if (rc == getPlayerId()) {
        /* Remote control flag says it's ours */
        if (ow == getPlayerId())
            return rc_Normal;
        else
            return rc_OurRemoteControlled;
    } else {
        /* Remote control flag says it's someone else's */
        if (ow == getPlayerId())
            return rc_RemoteControlled;
        else
            return rc_Other;
    }
}

/** Get new remote control state (end of turn).
    Returns the logical state after commands have been processed. */
TRemoteControlSetting
getNewRemoteControlState(const GShip& sh, GCommandContainer& cmds)
{
    // phost.pas::GetNewRCFlag
    using std::tolower;
    if (const GCommand* c = cmds.getCommand(GCommand::phc_RemoteControl, sh.getId())) {
        /* Check command */
        const string_t& arg = c->getArg();
        if (arg.size() > 0) {
            switch (tolower(uint8_t(arg[0]))) {
             case 'c': return rc_Applying;
             case 'a': return rc_Normal;
             case 'd': return rc_Dropping;
             default:  return rc_Forbidden;
            }
        } else {
            /* Invalid command */
            return getOldRemoteControlState(sh);
        }
    } else {
        /* No change */
        return getOldRemoteControlState(sh);
    }
}

/** Get indicator color for a remote control setting. */
TFrameColor
getRemoteControlColor(TRemoteControlSetting v)
{
    // phost.par::rc_Colors
    static const TFrameColor colors[] = {
        cf_Red,                 // rc_Forbidden
        cf_None,                // rc_Normal
        cf_Green,               // rc_RemoteControlled
        cf_Green,               // rc_Applying
        cf_Yellow,              // rc_Dropping
        cf_None,                // rc_Other
        cf_Red,                 // rc_OtherForbidden
        cf_None                 // rc_OurRemoteControlled
    };
    return colors[v];
}

/** Remote control. */
void
doRemoteControl(GShip& sh, GCommandContainer& cmds)
{
    /* Mapping of TRemoteControlSetting to question that can be used to leave this state. */
    static const char*const rc_questions[] = {
        N_("Allow remote control of this ship?"),
        N_("Forbid remote control of this ship?"),
        N_("Drop remote control of this ship?"),
        N_("Cancel request for remote control of this ship?"),
        N_("Cancel dropping remote control of this ship?"),
        N_("Request remote control of this ship?"),
        N_("Request remote control of this ship?"),
        N_("Forbid remote control of this ship?"),
    };

    if (UIMessageBox(_(rc_questions[getNewRemoteControlState(sh, cmds)]), _("Remote Control")).doYesNoDialog()) {
        int new_verb = rc_verb_index[getNewRemoteControlState(sh, cmds)] ^ 1;
        int old_verb = rc_verb_index[getOldRemoteControlState(sh)];

        if (new_verb == old_verb) {
            cmds.removeCommand(GCommand::phc_RemoteControl, sh.getId());
        } else {
            cmds.setCommand(GCommand::phc_RemoteControl, sh.getId(), rc_verbs[new_verb]);
        }

        sh.markDirty();
    }
}

/** Set remote control status to a particular value. This adds or
    removes a command as needed, without user interaction.

    \param sh Ship
    \param cmds Command container
    \param verb Target state
    \retval true A command has been given or removed to obtain the state
    \retval false Nothing changed because the state already was correct, or cannot be reached */
bool
setRemoteControl(GShip& sh, GCommandContainer& cmds, TRemoteControlVerb verb)
{
    // phost.pas::BringShipIntoRCState
    TRemoteControlSetting oldState = getOldRemoteControlState(sh);
    TRemoteControlSetting newState = getNewRemoteControlState(sh, cmds);

    bool result;

    if (rc_verb_index[newState] == verb) {
        /* I am already in the right state */
        result = false;
    } else if (rc_verb_index[oldState] == verb) {
        /* I was in the right state at the beginning of the turn */
        result = cmds.removeCommand(GCommand::phc_RemoteControl, sh.getId());
    } else if ((rc_verb_index[oldState] ^ 1) == verb) {
        /* I can reach the desired state by issuing a command */
        result = (cmds.setCommand(GCommand::phc_RemoteControl, sh.getId(), rc_verbs[verb]) != 0);
    } else {
        /* The state cannot be reached, e.g. "drop" for a ship I own. */
        result = false;
    }

    if (result) {
        sh.markDirty();
    }

    return result;
}
