/**
  *  \file game/objtypes.cc
  *  \brief Object types
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/objtypes.h"
#include "game/obj.h"
#include "cpluslib/assert.h"

/****************************** GObjectType ******************************/

GObjectType::~GObjectType()
{ }

/** Check emptiness.
    \return true this type is empty, i.e. has no objects
    \return false this type is not empty, i.e. has objects */
bool
GObjectType::isEmpty() const
{
    return findNextIndex(0) == 0;
}

/** Check unit type.
    \return true this type has precisely one object
    \return false this type has more or fewer than one objects */
bool
GObjectType::isUnit() const
{
    int n = findNextIndex(0);
    return n != 0 && findNextIndex(n) == 0;
}

/** Count objects.
    \return Number of objects */
int
GObjectType::countObjects() const
{
    int result = 0;
    for (int i = findNextIndex(0); i != 0; i = findNextIndex(i)) {
        ++result;
    }
    return result;
}

/** Find nearest object.
    \return Index of object, 0 if none */
int
GObjectType::findNearestIndex(GPoint pt)
{
    int found = 0;
    int32_t dist2 = 0;
    bool inside = false;

    for (int i = findNextIndex(0); i != 0; i = findNextIndex(i)) {
        if (GMapObject* mo = dynamic_cast<GMapObject*>(&getObjectByIndex(i))) {
            GCircularObject* co = dynamic_cast<GCircularObject*>(mo);
            int32_t ndist2 = distanceSquared(mo->getPos(), pt);
            bool ninside = (co != 0 && (ndist2 <= co->getRadiusSquared()));

            /* Pick this object, if
               - it's the first one
               - it's closer than our previous choice
               - we're inside it but not inside the previous one */
            if (found == 0
                || (inside == ninside && ndist2 < dist2)
                || (ninside && !inside))
            {
                found  = i;
                dist2  = ndist2;
                inside = ninside;
            }
        }
    }
    return found;
}

/** Get previous object before index, with wrap. If the last object of a kind is
    reached, search starts again at the beginning. Can filter marked objects.
    \param index Index to start search at.
    \param marked true to return only marked objects */
int
GObjectType::findPreviousIndexWrap(int index, bool marked)
{
    int n = findPreviousIndexNoWrap(index, marked);
    if (n == 0) {
        n = findPreviousIndexNoWrap(0, marked);
    }
    return n;
}

/** Get next object after index, with wrap. If the last object of a kind is
    reached, search starts again at the beginning. Can filter marked objects.
    \param index Index to start search at.
    \param marked true to return only marked objects */
int
GObjectType::findNextIndexWrap(int index, bool marked)
{
    int n = findNextIndexNoWrap(index, marked);
    if (n == 0) {
        n = findNextIndexNoWrap(0, marked);
    }
    return n;
}

/** Get previous object before index. Can filter marked objects.
    The returned object is guaranteed to exist.
    \param index Index to start search at.
    \param marked true to return only marked objects */
int
GObjectType::findPreviousIndexNoWrap(int index, bool marked)
{
    do {
        index = getPreviousIndex(index);
    } while (index != 0 && (!isValidIndex(index) || (marked && !getObjectByIndex(index).isMarked())));
    return index;
}

/** Get next object after index. Can filter marked objects.
    The returned object is guaranteed to exist.
    \param index Index to start search at.
    \param marked true to return only marked objects */
int
GObjectType::findNextIndexNoWrap(int index, bool marked)
{
    do {
        index = getNextIndex(index);
    } while (index != 0 && (!isValidIndex(index) || (marked && !getObjectByIndex(index).isMarked())));
    return index;
}

/** Get next object after index.
    This function is the same as findNextIndexNoWrap(index, false).
    Since it cannot filter objects, it can be a 'const' function.
    It is intended for iteration.
    \param index Index to start search at. */
int
GObjectType::findNextIndex(int index) const
{
    do {
        index = getNextIndex(index);
    } while (index != 0 && !isValidIndex(index));
    return index;
}

/**************************** GProxyObjectType ***************************/

GProxyObjectType::GProxyObjectType()
    : type(0)
{ }

GProxyObjectType::~GProxyObjectType()
{ }

GObject&
GProxyObjectType::getObjectByIndex(int index)
{
    /* This is not allowed to be called when isValidIndex() is false,
       so we can be sure that type is valid. */
    ASSERT(type != 0);
    return type->getObjectByIndex(index);
}

GUniverse*
GProxyObjectType::getUniverseByIndex(int index)
{
    ASSERT(type != 0);
    return type->getUniverseByIndex(index);
}

bool
GProxyObjectType::isValidIndex(int index) const
{
    return type != 0 && type->isValidIndex(index);
}

int
GProxyObjectType::getNextIndex(int index) const
{
    return type != 0 ? type->getNextIndex(index) : 0;
}

int
GProxyObjectType::getPreviousIndex(int index) const
{
    return type != 0 ? type->getPreviousIndex(index) : 0;
}

void
GProxyObjectType::setObjectType(GObjectType* type)
{
    this->type = type;
    if (type != 0) {
        conn_set_changed = type->sig_set_changed.add(&sig_set_changed, &Signal1<int>::raise);
    } else {
        conn_set_changed.disconnect();
    }
}

GObjectType*
GProxyObjectType::getObjectType()
{
    return type;
}

/*************************** GIndexedObjectType **************************/

GIndexedObjectType::GIndexedObjectType(int max)
    : max(max)
{ }

GIndexedObjectType::~GIndexedObjectType()
{ }

int
GIndexedObjectType::getNextIndex(int index) const
{
    return index < max ? index+1 : 0;
}

int
GIndexedObjectType::getPreviousIndex(int index) const
{
    return index > 0 ? index-1 : max;
}
