/**
  *  \file gfx/gfx.cc
  *  \brief Graphics Kernel
  *
  *  Implementation or graphics kernel (higher-level API wrapped
  *  around SDL).
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <iostream>
#include <SDL_error.h>
#include "gfx/gfx.h"
#include "util/except.h"
#include "cpluslib/assert.h"
#include "cpluslib/bitops.h"

/** \class GfxCanvas
    \brief Drawing area

    Offers an interface for simple graphics primitives. Descendants
    have to actually implement them. Based on the simple graphics
    primitives, we offer a few more complex ones in gfxcompl.h. The
    complex primitives use a GfxContext object to store the common
    parameters. This permits `stateful' programming. The simple
    graphics primitives are stateless.
    
    The general procedure for drawing something on a canvas is
    \code
       c->lock();
       // draw something
       c->unlock();
       c->update(...);     // update relevant parts
    \endcode
    We'll be going to hide that somewhere deep in the GUI class
    redraw management. */

/** Destructor. */
GfxCanvas::~GfxCanvas()
{ }

/** Check if rectangle is clipped. If it is not clipped, it is completely
    visible. */
bool
GfxCanvas::isClipped(GfxRect r)
{
    return r != computeClipRect(r);
}

/** Check if rectangle is visible. A rectangle is visible if at least
    one pixel of it can be seen. */
bool
GfxCanvas::isVisible(GfxRect r)
{
    return computeClipRect(r).exists();
}

//////////////////////////////////////////////////////////////////////

#include "gfx/gfx-modetr.h"

/** Pixel. Implementation of the `put pixel' primitive for all modes. */
template<class T>
inline void
doPixel(SDL_Surface* sfc, int x, int y, typename T::pix_type color)
{
    T::poke (T::add ((uint8*)sfc->pixels + y * sfc->pitch, x), color);
}

/** Pixel. Implementation of the `put pixel with alpha' primitive for
    all modes. */
template<class T>
inline void
doPixelAlpha(SDL_Surface* sfc, int x, int y, typename T::pix_type color,
             alpha_type alpha)
{
    uint8* p = T::add ((uint8*)sfc->pixels + y * sfc->pitch, x);
    T::poke (p, T::mix (sfc->format, T::peek(p), color, alpha));
}

/** Horizontal Line. Implementation of the `horizontal line' primitive
    for all modes. */
template<class T>
inline void
doHLine(SDL_Surface* sfc, int x1, int y1, int x2, typename T::pix_type color,
        uint8 pat)
{
    if (pat == 0) {
        /* nothing */
    } else if (pat == 255) {
        /* solid line */
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (x1 < x2) {
            T::poke(p, color);
            p = T::add(p, 1);
            ++x1;
        }
    } else {
        /* pattern line */
        uint8 mask = ror8 (0x80, x1);
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (x1 < x2) {
            if (mask & pat)
                T::poke(p, color);
            p = T::add(p, 1);
            ++x1;
            mask = ror8 (mask, 1);
        }
    }
}

/** Horizontal Line. Implementation of the `horizontal line with alpha
    channel' primitive for all modes. */
template<class T>
inline void
doHLineAlpha(SDL_Surface* sfc, int x1, int y1, int x2, typename T::pix_type color,
             uint8 pat, alpha_type alpha)
{
    if (pat == 0 || alpha == 0) {
        /* nothing */
    } else if (pat == 255) {
        /* solid line */
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (x1 < x2) {
            T::poke(p, T::mix(sfc->format, T::peek(p), color, alpha));
            p = T::add(p, 1);
            ++x1;
        }
    } else {
        /* pattern line */
        uint8 mask = ror8 (0x80, x1);
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (x1 < x2) {
            if (mask & pat)
                T::poke(p, T::mix(sfc->format, T::peek(p), color, alpha));
            p = T::add(p, 1);
            ++x1;
            mask = ror8 (mask, 1);
        }
    }
}

/** Vertical Line. Implementation of the `vertical line' primitive
    for all modes. */
template<class T>
inline void
doVLine(SDL_Surface* sfc, int x1, int y1, int y2, typename T::pix_type color,
        uint8 pat)
{
    if (pat == 0) {
        /* nothing */
    } else if (pat == 255) {
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (y1 < y2) {
            T::poke(p, color);
            p += sfc->pitch;
            ++y1;
        }
    } else {
        uint8 mask = ror8 (0x80, y1);
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (y1 < y2) {
            if (mask & pat)
                T::poke(p, color);
            p += sfc->pitch;
            ++y1;
            mask = ror8 (mask, 1);
        }
    }
}

/** Vertical Line. Implementation of the `vertical line with Alpha channel'
    primitive for all modes. */
template<class T>
inline void
doVLineAlpha(SDL_Surface* sfc, int x1, int y1, int y2, typename T::pix_type color,
             uint8 pat, alpha_type alpha)
{
    if (pat == 0 || alpha == 0) {
        /* nothing */
    } else if (pat == 255) {
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (y1 < y2) {
            T::poke(p, T::mix(sfc->format, T::peek(p), color, alpha));
            p += sfc->pitch;
            ++y1;
        }
    } else {
        uint8 mask = ror8 (0x80, y1);
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (y1 < y2) {
            if (mask & pat)
                T::poke(p, T::mix(sfc->format, T::peek(p), color, alpha));
            p += sfc->pitch;
            ++y1;
            mask = ror8 (mask, 1);
        }
    }
}

/** Implementation of the `blit pattern' primitive for transparent patterns.
    \param sfc             target surface
    \param rect            area to restrict drawing to; its left/top corner is the left/top
                           corner of our output
    \param pt              position of MSB of data[0]
    \param bytes_per_line  size of one line (no restrictions!)
    \param data            bitmap data
    \param color           color value in target format */
template<class T>
inline void
doBlitPatternTransp(SDL_Surface* sfc, GfxRect rect, const GfxPoint& pt,
                    int bytes_per_line, const uint8* data, typename T::pix_type color)
{
    int x = rect.x - pt.x;                // columns to skip in data image
    int y = rect.y - pt.y;                // lines to skip in data image
    data += bytes_per_line * y + x/8;

    uint8  zmask = 0x80 >> (x & 7);
    uint8* zmem = T::add ((uint8*)sfc->pixels + rect.y * sfc->pitch, rect.x);
    while (rect.h > 0) {
        uint8 mask = zmask;
        uint8* mem = zmem;
        const uint8* pat = data;
        for (unsigned k = 0; k < rect.w; ++k) {
            if (*pat & mask)
                T::poke (mem, color);
            mem = T::add(mem, 1);
            mask >>= 1;
            if (!mask) {
                mask = 0x80;
                ++pat;
            }
        }
        --rect.h;
        data += bytes_per_line;
        zmem += sfc->pitch;
    }
}

/** Implementation of the `blit pattern' primitive with background color.
    \param sfc             target surface
    \param rect            area to restrict drawing to; its left/top corner is the left/top
                           corner of our output
    \param pt              position of MSB of data[0]
    \param bytes_per_line  size of one line (no restrictions!)
    \param data            bitmap data
    \param color           color value in target format
    \param bg              background color in target format */
template<class T>
inline void
doBlitPattern(SDL_Surface* sfc, GfxRect rect, const GfxPoint& pt,
              int bytes_per_line, const uint8* data, typename T::pix_type color,
              typename T::pix_type bg)
{
    int x = rect.x - pt.x;                // columns to skip in data image
    int y = rect.y - pt.y;                // lines to skip in data image
    data += bytes_per_line * y + x/8;

    uint8  zmask = 0x80 >> (x & 7);
    uint8* zmem = T::add ((uint8*)sfc->pixels + rect.y * sfc->pitch, rect.x);
    while (rect.h > 0) {
        uint8 mask = zmask;
        uint8* mem = zmem;
        const uint8* pat = data;
        for (unsigned k = 0; k < rect.w; ++k) {
            if (*pat & mask)
                T::poke (mem, color);
            else
                T::poke (mem, bg);
            mem = T::add(mem, 1);
            mask >>= 1;
            if (!mask) {
                mask = 0x80;
                ++pat;
            }
        }
        --rect.h;
        data += bytes_per_line;
        zmem += sfc->pitch;
    }
}

/** Implementation of the `blit pattern' primitive with transparency.
    \param sfc             target surface
    \param rect            area to restrict drawing to; its left/top corner is the left/top
                           corner of our output
    \param pt              position of MSB of data[0]
    \param bytes_per_line  size of one line (no restrictions!)
    \param data            bitmap data
    \param color           color value in target format
    \param bg              background color in target format
    \param alpha           transpareny */
template<class T>
inline void
doBlitPatternAlpha(SDL_Surface* sfc, GfxRect rect, const GfxPoint& pt,
                   int bytes_per_line, const uint8* data, typename T::pix_type color,
                   uint32 bg, alpha_type alpha)
{
    int x = rect.x - pt.x;                // columns to skip in data image
    int y = rect.y - pt.y;                // lines to skip in data image
    data += bytes_per_line * y + x/8;

    uint8  zmask = 0x80 >> (x & 7);
    uint8* zmem = T::add ((uint8*)sfc->pixels + rect.y * sfc->pitch, rect.x);
    while (rect.h > 0) {
        uint8 mask = zmask;
        uint8* mem = zmem;
        const uint8* pat = data;
        for (unsigned k = 0; k < rect.w; ++k) {
            if (*pat & mask)
                T::poke (mem, T::mix (sfc->format, T::peek(mem), color, alpha));
            else if (bg != GFX_TRANSPARENT)
                T::poke (mem, T::mix (sfc->format, T::peek(mem), bg, alpha));
            mem = T::add(mem, 1);
            mask >>= 1;
            if (!mask) {
                mask = 0x80;
                ++pat;
            }
        }
        --rect.h;
        data += bytes_per_line;
        zmem += sfc->pitch;
    }
}

/** Implementation of the `read pixel' primitive for all modes.
    \param sfc   Surface
    \param x,y   Position to read */
template<class T>
inline typename T::pix_type
doPeek(SDL_Surface* sfc, int x, int y)
{
    return T::peek(T::add((uint8*) sfc->pixels + y * sfc->pitch, x));
}

//////////////////////////////////////////////////////////////////////

/** \class GfxSurface
    \brief SDL GfxSurface.
    This is a direct interface to an SDL \c GfxSurface. */

/** Use existing surface. We'll draw on the specified surface.
    If aowner is true, the surface will be deleted when this
    object is destroyed. The surface must be unlocked; noone
    else must play with the lock state.

    This is used to get a drawing surface for the screen after
    SDL_SetVideoMode(). */
GfxSurface::GfxSurface(GfxSystemSurface asfc, bool aowner)
    : sfc(asfc), owner(aowner), locked(0)
{
    // RETARGET: assumes GfxSystemSurface is SDL_Surface*
}

/** Memory GfxSurface. Creates a surface with the specified pixel
    format and the specified dimensions. */
GfxSurface::GfxSurface(int wi, int he, GfxSystemPixelFormat fmt)
    : sfc(SDL_CreateRGBSurface(SDL_SWSURFACE, wi, he,
                               fmt->BitsPerPixel,
                               fmt->Rmask, fmt->Gmask,
                               fmt->Bmask, fmt->Amask)),
      owner(true),
      locked(0)
{
    // RETARGET: assumes GfxSystemPixelFormat is SDL_PixelFormat*
    if (!sfc)
        throw GraphicsIOException(SDL_GetError());
}

GfxSurface::~GfxSurface()
{
    if (owner && sfc)
        SDL_FreeSurface(sfc);
}

void
GfxSurface::setSurface(SDL_Surface* asfc, bool aown)
{
    /* FIXME: lock count? for now, assume it's not locked. */
    ASSERT(!locked);

    if (owner)
        SDL_FreeSurface(sfc);
    sfc = asfc;
    owner = aown;
}

void
GfxSurface::drawHLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha)
{
    /* clipping */
    int x1 = pt.x;
    int y1 = pt.y;
    int x2 = x1 + npix;
    if (y1 < 0 || y1 >= sfc->h)
        return;
    if (x1 < 0)
        x1 = 0;
    if (x2 > sfc->w)
        x2 = sfc->w;
    if (x1 >= x2)
        return;

    if (alpha == 255)
        GFX_MODE_SWITCH(doHLine, (sfc, x1, y1, x2, color, pat));
    else
        GFX_MODE_SWITCH(doHLineAlpha, (sfc, x1, y1, x2, color, pat, alpha));
    return;
}

void
GfxSurface::drawVLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha)
{
    /* clipping */
    int x1 = pt.x;
    int y1 = pt.y;
    int y2 = y1 + npix;
    if (x1 < 0 || x1 >= sfc->w)
        return;
    if (y1 < 0)
        y1 = 0;
    if (y2 > sfc->h)
        y2 = sfc->h;
    if (y1 >= y2)
        return;

    if (alpha == 255)
        GFX_MODE_SWITCH(doVLine, (sfc, x1, y1, y2, color, pat));
    else
        GFX_MODE_SWITCH(doVLineAlpha, (sfc, x1, y1, y2, color, pat, alpha));
    return;
}

void
GfxSurface::drawPixel(const GfxPoint& pt, uint32 color, alpha_type alpha)
{
    if (pt.x < 0 || pt.x >= sfc->w || pt.y < 0 || pt.y >= sfc->h)
        return;

    if (alpha == 255)
        GFX_MODE_SWITCH(doPixel, (sfc, pt.x, pt.y, color));
    else
        GFX_MODE_SWITCH(doPixelAlpha, (sfc, pt.x, pt.y, color, alpha));
}

void
GfxSurface::drawBar(GfxRect rect, uint32 color, uint32 bg,
                    const GfxFillPattern& pat, alpha_type alpha)
{
    rect.clipSize(sfc->w, sfc->h);
    if (!rect.exists())
        return;

    if (alpha == 255 && pat.isBlack()) {
        SDL_FillRect(sfc, &rect, color);
    } else {
        int x1 = rect.x, y1 = rect.y, x2 = rect.x + rect.w, h = rect.h;
        if (alpha == 255) {
            if (bg == GFX_TRANSPARENT)
                while (h > 0) {
                    GFX_MODE_SWITCH (doHLine, (sfc, x1, y1, x2, color, pat[y1]));
                    ++y1; --h;
                }
            else
                while (h > 0) {
                    GFX_MODE_SWITCH (doHLine, (sfc, x1, y1, x2, color, pat[y1]));
                    GFX_MODE_SWITCH (doHLine, (sfc, x1, y1, x2, bg, ~pat[y1]));
                    ++y1; --h;
                }
        } else {
            while (h > 0) {
                GFX_MODE_SWITCH(doHLineAlpha, (sfc, x1, y1, x2, color, pat[y1], alpha));
                if (bg != GFX_TRANSPARENT)
                    GFX_MODE_SWITCH(doHLineAlpha, (sfc, x1, y1, x2, bg, ~pat[y1], alpha));
                ++y1;
                --h;
            }
        }
    }
}

/** Lock surface. Locks the surface if needed. */
bool
GfxSurface::lock()
{
    if (SDL_MUSTLOCK (sfc)) {
        if (SDL_LockSurface(sfc) == 0) {
            ++locked;
            return true;
        } else {
            return false;
        }
    } else {
        return true;
    }
}

/** Unlock surface. Unlocks the surface if needed. */
void
GfxSurface::unlock()
{
    if (SDL_MUSTLOCK (sfc)) {
        SDL_UnlockSurface(sfc);
        --locked;
    }
}

/** Copy other surface (pixmap).
    \param pt    anchor point. src's (0,0) is here;
    \param src   source surface;
    \param rect  clip rectangle, in src coordinates.

    `blitSurface(pt, src, GfxRect(src))' will display the whole
    surface. `blitSurface(pt, src, GfxRect(2, 3, 4, 5))' will
    display a 4x5 area from surface src, at (x+2, y+3) up to but not
    including (x+2+4, y+3+5). */
void
GfxSurface::blitSurface(const GfxPoint& pt, GfxCanvas& src, GfxRect rect)
{
    if (GfxSurface* s = dynamic_cast<GfxSurface*>(&src)) {
        SDL_Rect dstrect = { pt.x + rect.x, pt.y + rect.y, 0, 0 };
        int lockc = locked;
        while (locked)
            unlock();
        SDL_BlitSurface(s->sfc, &rect, sfc, &dstrect);
        while (locked < lockc)
            lock();
    } else {
        // FIXME: this does not work at all.
        // (a) no color transformation
        // (b) must offset by rect.x, rect.y
        std::cerr << "NOTE: attempt to blit a non-surface '"
                  << src.toString() << "' at " << pt << "\n";
        for (unsigned int yy = 0; yy < rect.h; ++yy) {
            for (unsigned int xx = 0; xx < rect.w; ++xx) {
                drawPixel(GfxPoint(pt.x + xx, pt.y + yy), src.getPixel(GfxPoint(rect.x + xx, rect.y + yy)), GFX_ALPHA_OPAQUE);
            }
        }
    }
}

/** Display pattern (monochrome bitmap).
    \param rect    position, screen coordinates;
    \param pt      position of bit 7, data[0], on the screen;
    \param bytes_per_line  bytes per line in data;
    \param data    pixel data. bit 7 = left, bit 0 = right (VGA bit order);
    \param color   color of "on" bits;
    \param bg      color of "off" bits, or GFX_TRANSPARENT;
    \param alpha   alpha channel.

    \pre pt.x <= rect.x, pt.y <= rect.y

    Note that this function has no constraints on the bytes_per_line
    parameter. bytes_per_line can also be zero or negative for
    interesting effects. */
void
GfxSurface::blitPattern(GfxRect rect, const GfxPoint& pt, int bytes_per_line,
                        const uint8* data, uint32 color, uint32 bg,
                        alpha_type alpha)
{
    rect.clipSize(sfc->w, sfc->h);
    if (rect.exists()) {
        if (alpha == 255) {
            if (bg == GFX_TRANSPARENT)
                GFX_MODE_SWITCH(doBlitPatternTransp, (sfc, rect, pt, bytes_per_line, data, color));
            else
                GFX_MODE_SWITCH(doBlitPattern, (sfc, rect, pt, bytes_per_line, data, color, bg));
        } else {
            GFX_MODE_SWITCH(doBlitPatternAlpha, (sfc, rect, pt, bytes_per_line, data, color, bg, alpha));
        }
    }
}

/** Update. Causes the specified region be redisplayed. */
void
GfxSurface::update(GfxRect r)
{
    r.clipSize(sfc->w, sfc->h);
    if (r.w && r.h)
        SDL_UpdateRect(sfc, r.x, r.y, r.w, r.h);
}

GfxSystemSurface
GfxSurface::getSystemSurface()
{
    return sfc;
}

GfxRect
GfxSurface::computeClipRect(GfxRect r)
{
    r.clipSize(sfc->w, sfc->h);
    return r;
}

bool
GfxSurface::computeOffset(GfxCanvas& other, GfxPoint& /*pt*/)
{
    return this == &other;
}

GfxPoint
GfxSurface::getSize()
{
    return GfxPoint(sfc->w, sfc->h);
}

int
GfxSurface::getBitsPerPixel()
{
    return sfc->format->BitsPerPixel;
}

uint32
GfxSurface::getPixel(const GfxPoint& pt)
{
    if (pt.x >= 0 && pt.y >= 0 && pt.x < sfc->w && pt.y < sfc->h)
        GFX_MODE_SWITCH(return doPeek, (sfc, pt.x, pt.y));
    return 0;
}

//////////////////////////////////////////////////////////////////////

GfxFilter::GfxFilter(GfxCanvas& parent)
    : parent(parent)
{ }

bool
GfxFilter::lock()
{
    return parent.lock();
}

void
GfxFilter::unlock()
{
    parent.unlock();
}

void
GfxFilter::update(GfxRect r)
{
    parent.update(r);
}

GfxPoint
GfxFilter::getSize()
{
    return parent.getSize();
}

int
GfxFilter::getBitsPerPixel()
{
    return parent.getBitsPerPixel();
}

uint32
GfxFilter::getPixel(const GfxPoint& pt)
{
    return parent.getPixel(pt);
}
