/**
  *  \file int/parseexpr.cc
  *  \brief Interpreter: Expression Parser
  *
  *  \todo PCC 1.x has "#n" in the Assignment production. If we want to make
  *  it part of regular expressions, it probably makes more sense to have it
  *  in the Primary production. On the other hand, the most sensible solution
  *  probably is to implement it in the respective routine taking "#"-ed
  *  parameters.
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/parseexpr.h"
#include "int/builtin.h"
#include "int/unary.h"
#include "int/binary.h"
#include "int/interror.h"

/** Constructor. */
IntExpressionParser::IntExpressionParser(IntTokenizer& tok)
    : tok(tok)
{ }

/** Parse expression. Parses a Sequence production.
    \return parsed expression tree. Caller is responsible for freeing it. */
IntExprNode*
IntExpressionParser::parse()
{
    parseSequence();
    return stack.extract_last();
}

/** Parse expression. Parses an Or-Expr production (=no assignment, no sequence).
    \return parsed expression tree. Caller is responsible for freeing it. */
IntExprNode*
IntExpressionParser::parseNA()
{
    parseOr();
    return stack.extract_last();
}

void
IntExpressionParser::parseSequence()
{
    // sequence ::= assignment
    //            | sequence ';' assignment
    parseAssignment();
    while (tok.checkAdvance(tok.tSemicolon)) {
        parseAssignment();
        makeBinary(new IntSequenceNode());
    }
}

void
IntExpressionParser::parseAssignment()
{
    // assignment ::= or-expr
    //              | or-expr ':=' assignment
    parseOr();
    if (tok.checkAdvance(tok.tAssign)) {
        parseAssignment();
        makeBinary(new IntAssignmentNode());
    }
}

void
IntExpressionParser::parseOr()
{
    // or-expr ::= and-expr
    //           | or-expr 'Or' and-expr
    //           | or-expr 'Xor' and-expr
    parseAnd();
    while (1) {
        if (tok.checkAdvance(tok.tOR)) {
            parseAnd();
            makeBinary(new IntLogicalNode(IntOpcode::jIfTrue, biOr));
        } else if (tok.checkAdvance(tok.tXOR)) {
            parseAnd();
            makeBinary(new IntLogicalNode(IntOpcode::jIfEmpty, biXor));
        } else {
            break;
        }
    }
}

void
IntExpressionParser::parseAnd()
{
    // and-expr ::= not-expr
    //            | and-expr 'And' not-expr
    parseNot();
    while (tok.checkAdvance(tok.tAND)) {
        parseNot();
        makeBinary(new IntLogicalNode(IntOpcode::jIfFalse, biAnd));
    }
}

void
IntExpressionParser::parseNot()
{
    // not-expr ::= comparison
    //            | 'Not' not-expr
    int n = 0;
    while (tok.checkAdvance(tok.tNOT))
        ++n;
    parseComparison();
    if (n > 0) {
        if (n & 1) {
            // Negation
            makeUnary(new IntSimpleExprNode(IntOpcode::maUnary, unNot));
        } else {
            // Cast to bool
            makeUnary(new IntSimpleExprNode(IntOpcode::maUnary, unBool));
        }
    }
}

void
IntExpressionParser::parseComparison()
{
    // comparison ::= concat-expr
    //              | comparison '=' concat-expr
    //              | comparison '<' concat-expr
    //              | comparison '>' concat-expr
    //              | comparison '<=' concat-expr
    //              | comparison '>=' concat-expr
    //              | comparison '<>' concat-expr
    parseConcat();
    while (1) {
        uint8_t mode;
        if (tok.checkAdvance(tok.tEQ)) {
            mode = biCompareEQ;
        } else if (tok.checkAdvance(tok.tLT)) {
            mode = biCompareLT;
        } else if (tok.checkAdvance(tok.tGT)) {
            mode = biCompareGT;
        } else if (tok.checkAdvance(tok.tLE)) {
            mode = biCompareLE;
        } else if (tok.checkAdvance(tok.tGE)) {
            mode = biCompareGE;
        } else if (tok.checkAdvance(tok.tNE)) {
            mode = biCompareNE;
        } else {
            break;
        }
        parseConcat();
        makeBinary(new IntCaseExprNode(mode));
    }
}

void
IntExpressionParser::parseConcat()
{
    // concat-expr ::= add-expr
    //               | concat-expr "#" add-expr
    //               | concat-expr "&" add-expr
    parseAdd();
    while (1) {
        uint8_t mode;
        if (tok.checkAdvance(tok.tHash)) {
            mode = biConcat;
        } else if (tok.checkAdvance(tok.tAmpersand)) {
            mode = biConcatEmpty;
        } else {
            break;
        }
        parseAdd();
        makeBinary(new IntSimpleExprNode(IntOpcode::maBinary, mode));
    }
}

void
IntExpressionParser::parseAdd()
{
    // add-expr ::= mult-expr
    //            | add-expr "+" mult-expr
    //            | add-expr "-" mult-expr
    parseMult();
    while (1) {
        uint8_t mode;
        if (tok.checkAdvance(tok.tPlus)) {
            mode = biAdd;
        } else if (tok.checkAdvance(tok.tMinus)) {
            mode = biSub;
        } else {
            break;
        }
        parseMult();
        makeBinary(new IntSimpleExprNode(IntOpcode::maBinary, mode));
    }
}

void
IntExpressionParser::parseMult()
{
    // mult-expr ::= neg-expr
    //             | mult-expr "*" neg-expr
    //             | mult-expr "/" neg-expr
    //             | mult-expr "\" neg-expr
    //             | mult-expr "Mod" neg-expr
    parseNeg();
    while (1) {
        uint8_t mode;
        if (tok.checkAdvance(tok.tMultiply)) {
            mode = biMult;
        } else if (tok.checkAdvance(tok.tSlash)) {
            mode = biDivide;
        } else if (tok.checkAdvance(tok.tBackslash)) {
            mode = biIntegerDivide;
        } else if (tok.checkAdvance(tok.tMOD)) {
            mode = biRemainder;
        } else {
            break;
        }
        parseNeg();
        makeBinary(new IntSimpleExprNode(IntOpcode::maBinary, mode));
    }
}

void
IntExpressionParser::parseNeg()
{
    // neg-expr ::= pow-expr
    //            | "-" neg-expr
    //            | "+" neg-expr
    bool neg = false;
    bool had = false;
    while (1) {
        if (tok.checkAdvance(tok.tMinus)) {
            neg = !neg;
            had = true;
        } else if (tok.checkAdvance(tok.tPlus)) {
            had = true;
        } else {
            break;
        }
    }

    if (tok.checkAdvance(tok.tNOT)) {
        /* This rule makes PCC accept "-not x". This isn't part of the
           original grammar, yet it's sensible in some way. Because it's
           rare, we don't optimize here. */
        parseNeg();
        makeUnary(new IntSimpleExprNode(IntOpcode::maUnary, unNot));
    } else {
        parsePow();
    }

    if (had) {
        if (neg) {
            makeUnary(new IntSimpleExprNode(IntOpcode::maUnary, unNeg));
        } else {
            makeUnary(new IntSimpleExprNode(IntOpcode::maUnary, unPos));
        }
    }
}

void
IntExpressionParser::parsePow()
{
    // pow-expr ::= primary-expr
    //            | primary-expr "^" neg-expr
    parsePrimary();
    if (tok.checkAdvance(tok.tCaret)) {
        parseNeg();
        makeBinary(new IntSimpleExprNode(IntOpcode::maBinary, biPow));
    }
}

/* primary-expr ::= "(" sequence ")"
                  | literal
                  | identifier {["(" arglist ")"] | ["." field]}*

   This grammar allows "a .b", but not "(a).b" for field references.
   Although inconsistent with other programming languages, this is
   actually a good thing, because it serves to disambiguate code like
   "Ship(sid).Name := 'xxx'" (which would otherwise be ambiguous to
   a procedure call with an assignment-expression as parameter:
   "Ship (sid .Name := 'xxx')" */
void
IntExpressionParser::parsePrimary()
{
    if (tok.checkAdvance(tok.tLParen)) {
        // Parenthesized expression
        parseSequence();
        if (!tok.checkAdvance(tok.tRParen))
            throw IntError::expectSymbol(")");
    } else if (tok.getCurrentToken() == tok.tInteger || tok.getCurrentToken() == tok.tBoolean) {
        // Integer literal
        IntLiteralExprNode* lit = new IntLiteralExprNode;
        stack.push_back(lit);
        lit->setValue(tok.getCurrentToken() == tok.tBoolean
                      ? makeBoolValue(tok.getCurrentInteger())
                      : makeIntValue(tok.getCurrentInteger()));
        tok.readNext();
    } else if (tok.getCurrentToken() == tok.tFloat) {
        // Float literal
        IntLiteralExprNode* lit = new IntLiteralExprNode;
        stack.push_back(lit);
        lit->setValue(makeFloatValue(tok.getCurrentFloat()));
        tok.readNext();
    } else if (tok.getCurrentToken() == tok.tString) {
        // String literal
        IntLiteralExprNode* lit = new IntLiteralExprNode;
        stack.push_back(lit);
        lit->setValue(new IntStringValue(tok.getCurrentString()));
        tok.readNext();
    } else if (tok.getCurrentToken() == tok.tIdentifier) {
        // Identifier
        string_t fname = tok.getCurrentString();

        // Special handling for builtin functions
        const TBuiltinFunctionDescriptor* bif;
        if (tok.readNext() == tok.tLParen && (bif = lookupBuiltinFunction(fname)) != 0) {
            // Builtin function
            tok.readNext();
            IntFunctionCallNode* fcn = bif->generator(*bif);
            stack.push_back(fcn);
            parseArglist(fcn);
            // Basic checks
            if (fcn->getNumArgs() < bif->min_args)
                throw IntError("Too few arguments for \"" + fname + "\""); // FIXME
            if (fcn->getNumArgs() > bif->max_args)
                throw IntError("Too many arguments for \"" + fname + "\""); // FIXME
        } else {
            // Regular function
            stack.push_back(new IntIdentifierNode(fname));
        }
        while (1) {
            if (tok.checkAdvance(tok.tLParen)) {
                // Array index / function call
                IntIndirectCallNode* fcn = new IntIndirectCallNode();
                fcn->setFunction(stack.extract_last());
                stack.push_back(fcn);
                parseArglist(fcn);
            } else if (tok.checkAdvance(tok.tDot) || tok.checkAdvance(tok.tArrow)) {
                // Member reference
                if (tok.getCurrentToken() != tok.tIdentifier)
                    throw IntError::expectIdentifier("field name");
                IntMemberNode* mem = new IntMemberNode(tok.getCurrentString());
                mem->setExpression(stack.extract_last());
                stack.push_back(mem);
                tok.readNext();
            } else {
                break;
            }
        }
    } else if (tok.checkAdvance(tok.tHash)) {
        // File number
        parsePrimary();
        makeUnary(new IntSimpleExprNode(IntOpcode::maUnary, unFileNr));
    } else if (tok.getCurrentToken() == tok.tEnd) {
        throw IntError("Expected operand");
    } else {
        throw IntError("Invalid expression");
    }
}

/** Parse argument list. Arguments will be read and added to \c fcn.
    \c fcn must already be on the stack, for exception safety.
    \param fcn Function call node to receive arguments */
void
IntExpressionParser::parseArglist(IntFunctionCallNode* fcn)
{
    if (tok.checkAdvance(tok.tRParen)) {
        // "foo()"
    } else {
        // "foo(args...)"
        do {
            parseSequence();
            fcn->addArgument(stack.extract_last());
        } while (tok.checkAdvance(tok.tComma));
        if (!tok.checkAdvance(tok.tRParen))
            throw IntError::expectSymbol(")");
    }
}

/** Make unary operation. Expects one tree node on the stack.
    \param n Tree node, will be pushed onto the stack */
void
IntExpressionParser::makeUnary(IntSimpleRValueExprNode* n)
{
    IntExprNode* a = stack.extract_last();
    n->setUnary(a);
    stack.push_back(n);
}

/** Make binary operation. Expects two tree nodes on the stack.
    \param n Tree node, will be pushed onto the stack */
void
IntExpressionParser::makeBinary(IntSimpleRValueExprNode* n)
{
    IntExprNode* b = stack.extract_last();
    IntExprNode* a = stack.extract_last();
    n->setBinary(a, b);
    stack.push_back(n);
}

/** Make ternary operation. Expects three tree nodes on the stack.
    \param n Tree node, will be pushed onto the stack */
void
IntExpressionParser::makeTernary(IntSimpleRValueExprNode* n)
{
    IntExprNode* c = stack.extract_last();
    IntExprNode* b = stack.extract_last();
    IntExprNode* a = stack.extract_last();
    n->setTernary(a, b, c);
    stack.push_back(n);
}
