/**
  *  \file tools/c2export.cc
  *  \brief "Export" utility
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <memory>
#include <cstring>
#include "arch/platform.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/config.h"
#include "game/coord.h"
#include "game/db.h"
#include "game/gamedirs.h"
#include "game/global.h"
#include "game/load.h"
#include "game/mission.h"
#include "game/msg.h"
#include "game/msgglobal.h"
#include "game/pref.h"
#include "game/racename.h"
#include "game/reg.h"
#include "game/sel.h"
#include "game/spec.h"
#include "game/stat.h"
#include "game/storage.h"
#include "game/storage/overview.h"
#include "game/team.h"
#include "int/compound.h"
#include "int/context.h"
#include "int/data.h"
#include "int/export.h"
#include "int/export/expdbf.h"
#include "int/export/expjson.h"
#include "int/export/expsep.h"
#include "int/export/exptext.h"
#include "int/if/conif.h"
#include "int/if/initif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "int/names.h"
#include "int/propenum.h"
#include "io/streams.h"
#include "io/textfile.h"
#include "util/answer.h"
#include "util/console.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "util/recode.h"
#include "version.h"

#define VERSION PCC2_VERSION

namespace {
    enum OutputType {
        tText,
        tTable,
        tCommaSV,
        tTabSV,
        tSemicolonSV,
        tJSON,
        tDBF
    };

    /* Console currently implements iostream, not Stream, but we need
       a Stream to use with TextFile. */
    class ConsoleWriter : public Stream {
     public:
        std::size_t read(char* /*buffer*/, std::size_t /*max*/)
            { return 0; }
        std::size_t write(const char* buffer, std::size_t max)
            {
                console << string_t(buffer, max);
                return max;
            }
        int getCapabilities() const
            { return C_WRITE; }
        void seek(fpos_t /*pos*/)
            { }
        fpos_t getPos()
            { return 0; }
        fpos_t getSize()
            { return 0; }
        string_t getName()
            { return _("Console"); }
    };

    /* Meta-context for generating field names. Used to implement '-F'. */
    class MetaContext : public IntContext,
                        public IntPropertyAcceptor
    {
     public:
        MetaContext();
        int32_t lookup(const IntNameQuery& name);
        void set(int32_t index, IntValue* value);
        IntValue* get(int32_t index);
        bool next();
        IntContext* clone();
        GObject* getObject();
        void enumProperties(IntPropertyAcceptor& acceptor);
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);
        void addProperty(const string_t& name, TIntTypeHint th);

     private:
        std::vector<string_t> names;
        std::vector<uint8_t> types;
        std::size_t position;
    };
}

static const TIntMapping meta_mapping[] = {
    { "ID",   0, 0, thInt },
    { "NAME", 1, 0, thString },
    { "TYPE", 2, 0, thString },
};

MetaContext::MetaContext()
    : names(),
      types(),
      position(0)
{ }

int32_t
MetaContext::lookup(const IntNameQuery& name)
{
    return lookupName(name, meta_mapping, countof(meta_mapping));
}

void
MetaContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
MetaContext::get(int32_t index)
{
    static const char*const typeNames[] = {
        "any",                  // thNone
        "bool",                 // thBool
        "int",                  // thInt
        "float",                // thFloat
        "string",               // thString
        "procedure",            // thProcedure
        "function",             // thFunction
        "array",                // thArray
    };

    if (position < names.size()) {
        switch (meta_mapping[index].index) {
         case 0:
            return makeIntValue(position);
         case 1:
            return makeStringValue(names[position]);
         case 2:
            if (types[position] < countof(typeNames)) {
                return makeStringValue(typeNames[types[position]]);
            } else {
                return 0;
            }
        }
    }
    return 0;
}

bool
MetaContext::next()
{
    if (position+1 < names.size()) {
        ++position;
        return true;
    } else {
        return false;
    }
}

IntContext*
MetaContext::clone()
{
    // This object is never cloned in c2export. When we make MetaContext a proper
    // type for use by scripts, we'd have to implement this.
    throw IntError("not clonable");
}

GObject*
MetaContext::getObject()
{
    return 0;
}

void
MetaContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(meta_mapping, countof(meta_mapping));
}

string_t
MetaContext::toString(bool /*readable*/)
{
    return "#<meta>";
}

void
MetaContext::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

void
MetaContext::addProperty(const string_t& name, TIntTypeHint th)
{
    names.push_back(name);
    types.push_back(th);
}


static const char* progname;

static void
die(string_t str)
{
    console.write(LOG_ERROR, format("%s: %s", progname, str));
    std::exit(1);
}

/** Exit with help message. */
static void
help()
{
    console << format(_("PCC2 Export v%s - (c) 2011-2016 Stefan Reuther\n"), VERSION)
            << format(_("\n"
                        "Usage:\n"
                        "  %s [-h]\n"
                        "  %$0s [-opts] [-f F@W...] [-S|-P|-A OBJECT] [-t TYPE] DIR [ROOT] PLAYER\n\n"
                        "Options:\n"
                        "  -C CHARSET      Select character set\n"
                        "  -f FIELD@WIDTH  Add field to report\n"
                        "  -S              Export ships (same as '-A SHIP')\n"
                        "  -P              Export planets (same as '-A PLANET')\n"
                        "  -A OBJECT       Export specified object type\n"
                        "  -t TYPE         Generate specified output format\n"
                        "  -o FILE         Set output file name (default: stdout)\n"
                        "  -O CHARSET      Select output file character set (default: UTF-8 or console)\n"
                        "  -F              Output list of fields instead of game data\n"
                        "\n"
                        "Types:\n"
                        "  dbf             dBASE file (needs '-o')\n"
                        "  text            simple text table, default\n"
                        "  table           boxy text table\n"
                        "  csv, tsv, ssv   comma/tab/semicolon-separated values\n"
                        "  json            JSON (JavaScript)\n"
                        "\n"
                        "Report bugs to <Streu@gmx.de>\n"),
                      progname);
    console.flush();
    std::exit(0);
}

static const char*
fetchArg(const char* opt, const char*& p, char**& argv)
{
    const char* result = p+1;
    if (!*result) {
        result = *++argv;
    }
    if (!result) {
        die(format(_("option '%s' needs an argument"), opt));
    }
    p = 0;
    return result;
}

// FIXME: this function is almost copied from main.cc
static void
loadGame(int player, int flags)
{
    // Construct storage instance
    if ((flags & GDirectoryOverview::have_Gamedir) != 0) {
        GGameDataStorage::instance = new GGameDirectoryStorage(player);
    } else {
        ASSERT((flags & GDirectoryOverview::have_RST) != 0);
        GGameDataStorage::instance = new GGameResultStorage(player, (flags & GDirectoryOverview::have_TRN) != 0);
    }

    // Load
    try {
        loadCommon(getCurrentTurn(), *game_file_dir, *spec_file_dir, LoadBoth, player);
        if (Ptr<Stream> cdb = game_file_dir->openFileNT(format("chart%d.cc", player), Stream::C_READ))
            loadChartDatabase(*cdb, getCurrentTurn(), true);
        GGameDataStorage::instance->load(getCurrentTurn());
        getCurrentTurn().postprocess(GObject::Playable);
    }
    catch (FileProblemException& fpe) {
        die(format(_("Error loading turn: %s: %s")) << fpe.getFileName() << fpe.what());
    }
    catch (std::exception& e) {
        die(format(_("Error loading turn: %s")) << e.what());
    }
}

static IntContext*
findArray(const char* name)
{
    // Look up name
    IntVariableNames::index_t i = global_property_names.lookup(strUCase(name));
    if (i == IntVariableNames::nil) {
        die(format(_("unknown object type '%s'"), name));
    }

    // Check for array
    IntCallableValue* cv = dynamic_cast<IntCallableValue*>(global_values[i]);
    if (!cv) {
        die(format(_("unknown object type '%s'"), name));
    }

    // Check for content
    try {
        IntContext* ctx = cv->makeFirstContext();
        if (!ctx) {
            die(format(_("this game does not contain any objects of type '%s'"), name));
        }
        return ctx;
    }
    catch (IntError& e) {
        // This happens when they do something like '-ACADD',
        // because CAdd refuses makeFirstContext() with an
        // IntError::typeError. No need to display the very
        // error message; it's not a known object type, period.
        die(format(_("unknown object type '%s'"), name));
        return 0;
    }
}

static void
doTextExport(OutputType typ, IntExportFieldList& job, IntContext* ctx, TextFile& tf)
{
    switch (typ) {
     case tText:
        IntTextExporter(tf, false).doExport(ctx, ConstantAnswerProvider::sayYes, job);
        break;
     case tTable:
        IntTextExporter(tf, true).doExport(ctx, ConstantAnswerProvider::sayYes, job);
        break;
     case tCommaSV:
        IntSeparatedTextExporter(tf, ',').doExport(ctx, ConstantAnswerProvider::sayYes, job);
        break;
     case tTabSV:
        IntSeparatedTextExporter(tf, '\t').doExport(ctx, ConstantAnswerProvider::sayYes, job);
        break;
     case tSemicolonSV:
        IntSeparatedTextExporter(tf, ';').doExport(ctx, ConstantAnswerProvider::sayYes, job);
        break;
     case tJSON:
        IntJsonTextExporter(tf).doExport(ctx, ConstantAnswerProvider::sayYes, job);
        break;
     case tDBF:
        /* handled outside */
        break;
    }
}


static const ModuleInitializer* c2export_modules[] = {
    mod_base_console,
    mod_config,
    mod_int,
    mod_pref,
    mod_coord,
    mod_missions,
    mod_msg,
    mod_msgglobal,
    mod_racenames,
    mod_reg,
    mod_sel,
    mod_spec,
    mod_stat,
    mod_team,
    mod_intgame_common,
    mod_intgame_console,
    0
};

int
main(int /*argc*/, char** argv)
{
    // Initialize. Suppress all informative messages which would clutter up stdout.
    console.setMinimumSeverity(LOG_WARN);
    performInitialisation(OnStartup, c2export_modules);
    console.setMinimumSeverity(LOG_DEBUG);

    argv = Platform::convertArgumentsToUtf8(argv);
    progname = *argv;

    // Parse args
    IntExportFieldList job;
    const char* arg_array = 0;
    const char* arg_gamedir = 0;
    const char* arg_rootdir = 0;
    const char* arg_ocs = 0;
    const char* arg_outfile = 0;
    OutputType arg_typ = tText;
    int arg_race = 0;
    bool opt_fields = false;
    int i;
    const char* pp;
    while (const char* p = *++argv) {
        if (*p == '-') {
            while (p && *++p) {
                switch (*p) {
                 case 'C':
                    {
                        CharacterSet cs(CharacterSet::getCharsetByName(fetchArg("-C", p, argv)));
                        if (!cs.isValid())
                            die(_("the specified character set is not known"));
                        setGameCharacterSet(cs);
                    }
                    break;
                 case 'f':
                    pp = fetchArg("-f", p, argv);
                    try {
                        job.addList(pp);
                    }
                    catch (IntError& e) {
                        die(format("'-f %s': %s", pp, e.what()));
                    }
                    break;
                 case 'F':
                    opt_fields = true;
                    break;
                 case 'S':
                    arg_array = "SHIP";
                    break;
                 case 'P':
                    arg_array = "PLANET";
                    break;
                 case 'A':
                    arg_array = fetchArg("-A", p, argv);
                    break;
                 case 't':
                    {
                        const char* t = fetchArg("-t", p, argv);
                        if (std::strcmp(t, "text") == 0) {
                            arg_typ = tText;
                        } else if (std::strcmp(t, "table") == 0) {
                            arg_typ = tTable;
                        } else if (std::strcmp(t, "dbf") == 0) {
                            arg_typ = tDBF;
                        } else if (std::strcmp(t, "csv") == 0) {
                            arg_typ = tCommaSV;
                        } else if (std::strcmp(t, "tsv") == 0) {
                            arg_typ = tTabSV;
                        } else if (std::strcmp(t, "ssv") == 0) {
                            arg_typ = tSemicolonSV;
                        } else if (std::strcmp(t, "json") == 0) {
                            arg_typ = tJSON;
                        } else {
                            die(_("invalid output format specified"));
                        }
                    }
                    break;
                 case 'o':
                    arg_outfile = fetchArg("-o", p, argv);
                    break;
                 case 'O':
                    arg_ocs = fetchArg("-O", p, argv);
                    if (!CharacterSet::isCharsetNameUtf8(arg_ocs) && !CharacterSet::getCharsetByName(arg_ocs).isValid()) {
                        die(_("the specified character set is not known"));
                    }
                    break;
                 case 'h':
                    help();
                    break;
                 default:
                    die(format(_("invalid option specified. Use `%s -h' for help"), progname));
                    break;
                }
            }
        } else if (arg_race == 0 && stringValue(p, i) && i > 0 && i <= NUM_PLAYERS) {
            arg_race = i;
        } else if (!arg_gamedir) {
            arg_gamedir = p;
        } else if (!arg_rootdir) {
            arg_rootdir = p;
        } else {
            die(_("too many arguments"));
        }
    }

    // Validate args
    if (!arg_array) {
        die(_("please specify the object type to export ('-P', '-S', '-A')"));
    }

    // Default field set
    if (job.size() == 0) {
        if (opt_fields) {
            job.addList("NAME@30,TYPE@10");
        } else {
            job.addList("ID@5,NAME@30");
        }
    }

    // Set up game directory
    initGameGlobals(arg_gamedir ? arg_gamedir : ".",
                    arg_rootdir ? string_t(arg_rootdir) : Platform::getDataDirectory());

    // Check for game data
    GDirectoryOverview ov;
    ov.scan(*game_file_dir);
    if (arg_race != 0) {
        if ((ov.getPlayerFlags(arg_race) & (GDirectoryOverview::have_RST | GDirectoryOverview::have_Gamedir)) == 0) {
            die(format(_("no game data available for player %d"), arg_race));
        }
    } else {
        arg_race = ov.getDefaultPlayer();
        if (!arg_race) {
            if (ov.getGlobalFlags() == 0) {
                die(_("no game data found"));
            } else {
                die(_("please specify the player number"));
            }
        }
    }
    host = ov.getHostVersion(arg_race);

    // Load game data with suppressed log messages.
    console.setMinimumSeverity(LOG_WARN);
    setPlayerIds(arg_race, arg_race);
    performInitialisation(OnEnterDirectory | OnEnterGame, c2export_modules);
    loadGame(arg_race, ov.getPlayerFlags(arg_race));
    setDisplayedTurn(0);
    performInitialisation(OnEnteredGame, c2export_modules);
    console.setMinimumSeverity(LOG_DEBUG);

    // What do we want to export?
    std::auto_ptr<IntContext> array(findArray(arg_array));
    if (opt_fields) {
        std::auto_ptr<MetaContext> meta(new MetaContext());
        array->enumProperties(*meta);
        array = meta;
    }

    // Do it.
    try {
        if (arg_typ == tDBF) {
            // Output to DBF file. Requires file name.
            if (arg_outfile == 0) {
                die(_("output to DBF file needs an output file name ('-o')"));
            }
            Ptr<Stream> s = Stream::openFile(arg_outfile, Stream::C_CREATE);
            IntDbfExporter(*s).doExport(array.get(), ConstantAnswerProvider::sayYes, job);
        } else {
            if (arg_outfile == 0) {
                // Output to console. Note that the console performs
                // character set conversion and adds proper newlines,
                // so we must tell TextFile to pass through our data
                // unmodified (UTF-8, no system newlines).
                if (arg_ocs != 0) {
                    console.write(LOG_WARN, _("WARNING: Option '-O' has been ignored because standard output is being used."));
                }
                ConsoleWriter cw;
                TextFile tf(cw);
                tf.setSystemNewline(false);
                tf.setUtf8();
                doTextExport(arg_typ, job, array.get(), tf);
                tf.sync();
            } else {
                // Output to file
                Ptr<Stream> s = Stream::openFile(arg_outfile, Stream::C_CREATE);
                TextFile tf(*s);
                if (arg_ocs != 0) {
                    if (CharacterSet::isCharsetNameUtf8(arg_ocs)) {
                        tf.setUtf8();
                    } else {
                        tf.setCharacterSet(CharacterSet::getCharsetByName(arg_ocs));
                    }
                }
                doTextExport(arg_typ, job, array.get(), tf);
                tf.sync();
            }
        }
    }
    catch (FileProblemException& fpe) {
        die(format("%s: %s", fpe.getFileName(), fpe.what()));
    }
    catch (IntError& ie) {
        die(ie.what());
    }
    catch (GError& ge) {
        die(ge.getUserError());
    }
    catch (std::exception& e) {
        die(e.what());
    }
    return 0;
}
