/**
  *  \file ui/richdocument.h
  *  \brief Rich Document Display
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_RICHDOCUMENT_H
#define PCC_V2_UI_RICHDOCUMENT_H

#include <cstddef>
#include <memory>
#include "ui/richtext.h"
#include "cpluslib/ptrvector.h"
#include "ui/ui.h"
#include "cpluslib/sigslot.h"
#include "ui/listbox.h"
#include "ui/richobj.h"

class GfxContext;

/** Rich Document. This class allows formatting of RichText. It breaks
    lines according to margins, and adjusts line heights according to
    used fonts.

    In addition to text, a RichDocument allows usage of block objects,
    which are instances of RichDocumentBlockObject, and can appear as
    floating block objects at the left or right side of the document,
    or as centered block objects interrupting the text flow. The main
    use for this feature is images.

    To use, configure the width and margin as needed, and add text
    using add(), addNewline(), etc., in any order you wish. This will
    accumulate all the text in an internal buffer of Item's
    (\c content); partially finished items will remain in an invisible
    intermediate buffer (\c last_chunk). To finish the document, call
    finish(). Afterwards, the RichDocument can be used for display. */
class RichDocument {
 public:
    struct Item {
        /* These values are sorted such that one can test '& Link' for link text */
        enum Kind {
            Normal     = 0,
            LinkTarget = 1,
            Link       = 0x80,
            LinkHover  = 0x81,
            LinkFocus  = 0x82
        };
        Kind kind;
        int x;
        int y;
        int w;           /* display width (0 for LinkTarget). */
        int font;
        UIColor::Color color;
        bool underline;
        bool key;
        string_t text;

        Item(Kind kind, int x, int y, int w, int font, UIColor::Color color, bool underline, bool key, const string_t& text)
            : kind(kind), x(x), y(y), w(w),
              font(font), color(color),
              underline(underline), key(key), text(text)
            { }
    };
    struct BlockObject {
        enum Kind {
            Right,
            Left,
            Center
        };
        Kind kind;
        GfxRect pos;
        std::auto_ptr<RichDocumentBlockObject> obj;

        BlockObject(Kind k, std::auto_ptr<RichDocumentBlockObject> obj)
            : kind(k),
              pos(0, 0, obj->getSize()),
              obj(obj)
            { }
    };

    /** Link id. To users, this is an opaque value which uniquely identifies a link.
        Internally, it is the index into \c content pointing to the LinkTarget item. */
    typedef uint32_t linkid_t;
    static const linkid_t nil = linkid_t(-1);

    RichDocument();
    ~RichDocument();

    void clear();
    void setPageWidth(int width);
    void setLeftMargin(int lm);
    void setRightMargin(int rm);
    int getPageWidth() const;
    int getLeftMargin() const;
    int getRightMargin() const;

    void add(const RichText& text);
    void add(const string_t& text);
    void add(const char* text);
    void addWordSeparator();
    void addNewline();
    void addParagraph();
    void addAt(int x, const RichText& text);
    void addRight(int x, const RichText& text);
    void addCentered(int x, const RichText& text);
    void addPreformatted(const RichText& text);
    void addFloatObject(std::auto_ptr<RichDocumentBlockObject> obj, bool left);
    void addCenterObject(std::auto_ptr<RichDocumentBlockObject> obj);
    void tabTo(int x);
    void finish();
    int getDocumentHeight() const;
    int getDocumentWidth() const;

    void draw(GfxContext& ctx, GfxRect area, int skipY);

    linkid_t getLinkFromPos(GfxPoint pt) const;
    string_t getLinkTarget(linkid_t link) const;
    void setLinkKind(linkid_t link, Item::Kind kind);

    linkid_t getNextLink(linkid_t id) const;
    linkid_t getNextLink(linkid_t id, GfxRect limit) const;
    linkid_t getPreviousLink(linkid_t id) const;
    linkid_t getPreviousLink(linkid_t id, GfxRect limit) const;
    bool isLinkVisible(linkid_t id, GfxRect limit) const;

    /* Options */
    enum {
        /* If set, paragraph breaks are full lines (default: half) */
        FullLinesBetweenParagraphs = 1
    };
    void setRenderOptions(int opts);
    int getRenderOptions() const;

 private:
    /* Up to first_this_line: completely rendered content.
       After first_this_line: x positions have been assigned, y positions have not been assigned yet */
    ptr_vector<Item> content;
    /* Contains a partial line with no positions assigned.
       If it contains spaces, it is guaranteed to fit into a
       line after the current x. */
    ptr_vector<Item> last_chunk;
    /* Contains all block objects. */
    ptr_vector<BlockObject> block_objs;

    /* Render options */
    int render_options;

    int x;
    int y;
    std::size_t first_this_line;

    int page_width;
    int left_margin;
    int right_margin;

    /* Current block object status. Indexed by Left/Right.
       - bo_index: index of last unprocessed block_objs entry. Can point
         past the end of block_objs if all objects of that type have been
         processed.
       - bo_width: width of current block object placeholder
       - bo_height: remaining height of current block object placeholder.
         Nonzero defines the object as being partially placed. */
    uint32_t bo_index[2];
    int bo_width[2];
    int bo_height[2];

    void process();
    void addY(int dy);
    void findNextObject(int side);
    void startNextObject(int side);
    void flushLine();
    void flushWord();
    void flushItems(std::size_t n);
};

/** Rich document widget. Displays a RichDocument and allows the user
    to interact with it. Since the RichDocument stores preformatted text,
    there are two ways to fill the document:
    - after doing layout, but before showing the widget, populate the
      document normally
    - while the widget is active, manipulate the document and finally
      call handleDocumentUpdate(). */
class UIRichDocument : public UIScrollable {
 public:
    enum {
        fl_Tab    = 1,
        fl_Return = 2,
        fl_Space  = 4,
        fl_Scroll = 8,

        /* Combination: single-page hypertext */
        fl_SingleHyper = fl_Tab + fl_Return + fl_Space,

        /* Combination: help viewer */
        fl_Help = fl_Tab + fl_Return + fl_Scroll
    };
    UIRichDocument(GfxPoint pref_size, uint16_t key_flags);

    RichDocument& getDocument();

    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
    void onResize(GfxRect oldSize, bool byLayout);
    bool handleEvent(const UIEvent& event, bool second_pass);
    void onStateChange(int state, bool enabled);
    void handleDocumentUpdate();
    void adjustToDocumentSize();

    RichDocument::linkid_t getSelectedLink() const;
    void setSelectedLink(RichDocument::linkid_t link);
    int getTopY() const;

    // UIScrollable
    void scrollTo(int pos, int corr);
    bool isAccessible(int pos);

    Signal1<string_t> sig_link_click;

 private:
    void setHoverLink(RichDocument::linkid_t link);
    void setLink(RichDocument::linkid_t link);
    int getScrollStep() const;
    void updateScrollable();

    RichDocument doc;
    GfxPoint pref_size;

    uint16_t key_flags;
    bool mdown;
    RichDocument::linkid_t selected_link;
    RichDocument::linkid_t hover_link;
};


/**************************** Inline Functions ***************************/

/** Set this document's page width.
    \param width Page width in pixels */
inline void
RichDocument::setPageWidth(int width)
{
    page_width = width;
}

/** Set this document's right margin.
    \param lm Right margin in pixels. */
inline void
RichDocument::setRightMargin(int rm)
{
    right_margin = rm;
}

/** Get page width. */
inline int
RichDocument::getPageWidth() const
{
    return page_width;
}

/** Get this document's current left margin. */
inline int
RichDocument::getLeftMargin() const
{
    return left_margin;
}

/** Get this document's current right margin. */
inline int
RichDocument::getRightMargin() const
{
    return right_margin;
}

/** Add zero-width word separator. Flushes the current word and allows
    a new line to be begun. */
inline void
RichDocument::addWordSeparator()
{
    flushWord();
}

/** Get height of document. */
inline int
RichDocument::getDocumentHeight() const
{
    return y;
}

/** Set rendering options. */
inline void
RichDocument::setRenderOptions(int opts)
{
    render_options = opts;
}

/** Get rendering options. */
inline int
RichDocument::getRenderOptions() const
{
    return render_options;
}


inline RichDocument&
UIRichDocument::getDocument()
{
    return doc;
}

inline void
UIRichDocument::adjustToDocumentSize()
{
    pref_size.y = doc.getDocumentHeight();
}

inline RichDocument::linkid_t
UIRichDocument::getSelectedLink() const
{
    return selected_link;
}



#endif
